/*
 *	cutsrt Ver.0.01		Copyright (C) 2009  kt
 *	cutass Ver.0.02		(cutsrt mod)
 */

#include <ctype.h>
#include <stdio.h>
#include <vector>


typedef enum {
	CUT_KEY_A,
	CUT_KEY_1,
	CUT_KEY_2,
	CUT_KEY_3,
	CUT_AVS,
	CUT_VCF,
	CUT_DEL
} cut_list_type;

bool debug;
static unsigned int fps;
std::vector<int> dellist;

int load_dellist(const char *filename, cut_list_type cut_type);
double get_cut_frame_number(double frame, int mode);
void cut_ass(const char *file_in, const char *file_out);
void cut_srt(const char *file_in, const char *file_out);

double time2frame(double time);
double frame2time(double frame);

void usage()
{
	char *msg =
		"cutsrt Ver.0.01  Copyright (C) 2009  kt\n"
		"cutass Ver.0.02  (cutsrt mod  2011)\n"
		"\n"
		"usage:\n"
		"  cutass [<options>] <input{.ass|.srt}>\n"
		"\n"
		"options:\n"
		"  -l <AviUtl_DelList.txt>\tdefault: input_del.txt\n"
		"  -t <trimfile.avs>\n"
		"  -v <vcffile.vcf>\n"
		"  -k{1|2|3} <TMPGEnc.keyframe>\n"
//		"  -i <input.ass|.srt>\n"
		"  -o <output{.ass|.srt}>\tdefault: input_new{.ass|.srt}\n"
		"  -fps <frame_rate_value>\tdefault: 30\n"
		"  -d  :show debug info\n"
	;
	fputs(msg, stderr);
}

int main(int argc, char *argv[])
{
	char buf[_MAX_PATH], buf2[_MAX_PATH];
	char *listfile = NULL, *infile = NULL, *outfile = NULL;
	cut_list_type	cut_type;
	fps = 30;
	debug = false;
	
	while (*++argv) {
		char c = (*argv)[0];
		if ((c == '-') || (c == '/')) {
			switch (tolower((*argv)[1])) {
			case 'f':
				if(strcmp((*argv),"-fps") == 0) {
					fps = atoi(*++argv);
					if (fps == 0) {
						fps = 30;
					}
				}
				break;
			case 'k':
				if(strcmp((*argv),"-k1") == 0) {
					listfile = *++argv;
					cut_type = CUT_KEY_1;
				} else 
				if(strcmp((*argv),"-k2") == 0) {
					listfile = *++argv;
					cut_type = CUT_KEY_2;
				} else 
				if(strcmp((*argv),"-k3") == 0) {
					listfile = *++argv;
					cut_type = CUT_KEY_3;
				} else {
					listfile = *++argv;
					cut_type = CUT_KEY_A;
				}
				break;
			case 't':
				listfile = *++argv;
				cut_type = CUT_AVS;
				break;
			case 'v':
				listfile = *++argv;
				cut_type = CUT_VCF;
				break;
			case 'l':
				listfile = *++argv;
				cut_type = CUT_DEL;
				break;
			case 'o':
				outfile = *++argv;
				break;
			case 'd':
				debug = true;
				break;
			case 'h':
				usage();
				break;
			default:
				break;
			}
			if (*argv == NULL) {
				usage();
				break;
			}
		} else {
			infile = *argv;
			
			char fin_path[_MAX_PATH];
			char drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], fext[_MAX_EXT];
			_splitpath(infile, drive, dir, fname, fext);
			_makepath(fin_path, drive, dir, fname, NULL);
			
			if (listfile == NULL) {
				listfile = buf;
				_makepath(listfile, drive, dir, fname, NULL);
				strcat(listfile, "_del.txt");
				cut_type = CUT_DEL;
			}
			if (outfile == NULL) {
				outfile = buf2;
				_makepath(outfile, drive, dir, fname, NULL);
				strcat(outfile, "_new");
			} else {
				_splitpath(outfile, drive, dir, fname, NULL);
				outfile = buf2;
				_makepath(outfile, drive, dir, fname, NULL);
			}
			
			if (load_dellist(listfile, cut_type) >= 0) {
				if (strcmp(fext,".ass")==0) {
					cut_ass(fin_path, outfile);
				} else 
				if (strcmp(fext,".srt")==0) {
					cut_srt(fin_path, outfile);
				} else 
//				if (strcmp(fext,"")==0)
				{
					cut_ass(fin_path, outfile);
					cut_srt(fin_path, outfile);
				}
			}
			listfile = NULL;
			outfile = NULL;
			fps = 30;
			debug = false;
		}
	}
	if (infile == NULL) {
		usage();
	}
	return 0;
}

int load_dellist(const char *filename, cut_list_type cut_type)
{
	dellist.clear();
	
	FILE *fp = fopen(filename, "rt");
	if (fp == NULL) {
		fprintf(stderr, "list file open error: %s\n", filename);
		return(-1);
	}
	char buf[256];
	int result = 0;
	dellist.push_back(-1);			// sentinel		mod move

	int start = 0, end = 0;
	int i=0, j, c;
	int trim_pos[2];
	fpos_t fpos;
	int cutframes = 0, outframes = 0;		// for debug

	if (cut_type == CUT_KEY_A) {
		int odd_frame = 0;
		int even_frame = 0;
		if (fgets(buf, sizeof(buf), fp) != NULL) {
			trim_pos[i] = atoi(buf);  i++;
			while (fgets(buf, sizeof(buf), fp) != NULL) {
				trim_pos[i] = atoi(buf);
				i ^= 1;
				if (i==0) {
					odd_frame += trim_pos[1] - trim_pos[0];
				} else {
					even_frame += (trim_pos[0]-1) - (trim_pos[1]+1);
				}
			}
			if ( odd_frame <= even_frame) {
				cut_type = CUT_KEY_1;
			} else {
				cut_type = CUT_KEY_2;
			}
if (debug) {
			fprintf(stderr, " [key]\to:%d\te:%d\t >> auto select:%d\n", odd_frame, even_frame, cut_type);
}
		}
		i = 0;
		fseek(fp, 0, SEEK_SET);
	}

	char *	fread_result;
	switch( cut_type ) {
	case CUT_KEY_A:
		result = -1;
		fprintf(stderr, " [key]\tAuto: Err!! Not select Odd/Even.\n\n");
		break;
	case CUT_KEY_1:
if (debug) {
		fprintf(stderr, " [key]\t1 : Odd\n");
}
		if (fgets(buf, sizeof(buf), fp) != NULL) {
			while (fgets(buf, sizeof(buf), fp) != NULL) {
				trim_pos[i] = atoi(buf);
				i ^= 1;
				if (i==0) {
if (debug) {
					cutframes -= end;
}
					start = trim_pos[0];  end = trim_pos[1];
					dellist.push_back(start);
					dellist.push_back(end);
if (debug) {
					fprintf(stderr, "\t  cut:\ts=%d\te=%d\n", start+1, end-1);
					outframes += (end - start -1);
					cutframes += (start +1);
}
				}
			}
		}
if (debug) {
		fprintf(stderr, "\n");
}
		break;
	case CUT_KEY_2:
if (debug) {
		fprintf(stderr, " [key]\t2 : Even\n");
}
		if (fgets(buf, sizeof(buf), fp) != NULL) {
			trim_pos[i] = atoi(buf) - 1;	i++;
			while (fgets(buf, sizeof(buf), fp) != NULL) {
				trim_pos[i] = atoi(buf);
				i ^= 1;
				if (i==0) {
if (debug) {
					cutframes -= end;
}
					start = trim_pos[0];  end = trim_pos[1];
					dellist.push_back(start);
					dellist.push_back(end);
if (debug) {
					fprintf(stderr, "\t  cut:\ts=%d\te=%d\n", start+1, end-1);
					outframes += (end - start -1);
					cutframes += (start +1);
}
				}
			}
		}
if (debug) {
		fprintf(stderr, "\n");
}
		break;
	case CUT_KEY_3:
if (debug) {
		end = -1;
		fprintf(stderr, " [key]\t3\n");
}
		while (fgets(buf, sizeof(buf), fp) != NULL) {
			trim_pos[i] = atoi(buf);
			i ^= 1;
			if (i==0) {
if (debug) {
				cutframes -= end;
}
				start = trim_pos[0];  end = trim_pos[1];
				dellist.push_back(start - 1);
				dellist.push_back(end + 1);
if (debug) {
				fprintf(stderr, "\t  cut:\ts=%d\te=%d\n", start, end);
				outframes += (end - start +1);
				cutframes += (start -1);
}
			}
		}
if (debug) {
		fprintf(stderr, "\n");
}
		break;
	case CUT_AVS:
if (debug) {
		end = -1;
		fprintf(stderr, " [avs]\n");
}
		while ((c = fgetc(fp)) !=EOF) {
			if (c == '#') {
				if (fgets(buf, sizeof(buf), fp) == NULL) {
					break;
				}
			} else 
			if (c == '/') {
				if ((c = fgetc(fp)) !=EOF) {
					if (c == '*') {
						while ((c = fgetc(fp)) !=EOF) {
							if (c =='*') {
								if ((c = fgetc(fp)) !=EOF) {
									if (c == '/') {
										break;
									}
								}
							}
						}
					}
				}
			} else 
			if (c == 'T') {
				fgetpos(fp, &fpos);
				if (3 == fread(buf,sizeof(char),3,fp)) {
					buf[3]='\0';
					if (strcmp("rim",buf) == 0) {
						i=0; j=0;
						while ((c = fgetc(fp)) !=EOF) {
							if (c == '(') {
								while ((c = fgetc(fp)) !=EOF) {
									if (('0'<=c && c<='9') || c =='-') {
										buf[i] = c;  i++;
									} else 
									if (c==',') {
										buf[i] = '\0'; trim_pos[j] = atoi(buf);
										i=0; j++;
									} else 
									if (c==')') {
if (debug) {
										cutframes -= end;
}
										buf[i] = '\0'; trim_pos[j] = atoi(buf);
										start = trim_pos[0];  end = trim_pos[1];
										end = (end<0) ? start-end-1 : end;
										dellist.push_back(start - 1);
										dellist.push_back(end + 1);
if (debug) {
										fprintf(stderr, "\t Trim:\ts=%d\te=%d\n", start, end);
										outframes += (end - start +1);
										cutframes += (start -1);
}
										break;
									}
								}
								break;
							}
						}
					} else {
						fsetpos(fp, &fpos);
					}
				}
			}
		}
if (debug) {
		fprintf(stderr, "\n");
}
		break;
	case CUT_VCF:
if (debug) {
		fprintf(stderr, " [vcf]\n");
}
		while (fgets(buf, sizeof(buf), fp) != NULL) {
if (debug) {
			cutframes -= (start+end);
}
			if (sscanf(buf, "VirtualDub.subset.AddRange(%d,%d);", &start, &end) == 2) {
				dellist.push_back(start - 1);
				dellist.push_back(start + end);
if (debug) {
				fprintf(stderr, "\tRange:\ts=%d\te=%d\n", start, start+end-1);
				outframes += end;
				cutframes += start;
}
			}
		}
if (debug) {
		cutframes += (start+end);
		fprintf(stderr, "\n");
}
		break;
	case CUT_DEL:
		start = -1;  end = -1;
		while ((fread_result = fgets(buf, sizeof(buf), fp)) != NULL) {
			if (buf[0] == '#')	continue;
			if (sscanf(buf, "%d * %d", &start, &end) == 2) {
				break;
			}
		}
		if (fread_result == NULL || start == -1 || end == -1) {
			fprintf(stderr, "list file read error\n");
			result = -1;
			break;
		}
if (debug) {
		fprintf(stderr, " [del]\n");
		fprintf(stderr, "\t list:\ts=%d", start);
		outframes = end+1;
		cutframes = start;
}
//		dellist.push_back(-1);			// sentinel		mod move
		dellist.push_back(start - 1);
		while (fgets(buf, sizeof(buf), fp) != NULL) {
			int del_start=-1, del_end=-1;
			if (buf[0] == '#')	continue;
			if (sscanf(buf, "%d * %d", &del_start, &del_end) == 2)	continue;	// skip: duplicate "*"
			if (sscanf(buf, "%d - %d", &del_start, &del_end) > 0) {
				if (del_end==-1)		del_end = del_start;		// case: 1frame cut
				dellist.push_back(del_start);
				dellist.push_back(del_end);
if (debug) {
				fprintf(stderr, "\te=%d\n", del_start-1);
				fprintf(stderr, "\t list:\ts=%d", del_end+1);
				cutframes += (del_end - del_start +1);
}
			}
		}
		dellist.push_back(end + 1);
if (debug) {
		fprintf(stderr, "\te=%d\n", end);
		fprintf(stderr, "\n");
		outframes -= cutframes;
}
		break;
	default:
		result = -1;
		break;
	}
	dellist.push_back(-1);			// sentinel
	dellist.push_back(-1);			// sentinel		mod fix(pre)
	fclose(fp);

	if (debug) {
		int t_time = frame2time(outframes);
		int c_time = frame2time(cutframes);
		int th, tm, ts, tms;
		int ch, cm, cs, cms;

		tms = t_time % 1000;
		t_time /= 1000;
		ts = t_time % 60;
		t_time /= 60;
		tm = t_time % 60;
		t_time /= 60;
		th = t_time % 60;

		cms = c_time % 1000;
		c_time /= 1000;
		cs = c_time % 60;
		c_time /= 60;
		cm = c_time % 60;
		c_time /= 60;
		ch = c_time % 60;

		fprintf(stderr, "\tTotal  frame: %d\ttime: %02d:%02d:%02d.%03d\n", outframes, th, tm, ts, tms);
		fprintf(stderr, "\t cut   frame: %d\ttime: %02d:%02d:%02d.%03d\n", cutframes, ch, cm, cs, cms);
		fprintf(stderr, "\n");
	}
	return(result);
}

double get_cut_frame_number(double frame, int mode)
{
	int cutframes = -1;
	for (int i = 0; i < dellist.size() - 1; i += 2) {
//		if (dellist[i+1] < 0) { break; }	// mod fix
		int oldcutframes = cutframes;
		cutframes += dellist[i+1] - dellist[i] + 1;
		if ((dellist[i] <= frame) && (frame <= dellist[i+1])) {
			// deleted
			if (mode > 0) {
				return dellist[i+1] + 1 - cutframes;
			} else if (mode < 0) {
				return dellist[i] - 1 - oldcutframes;
			} else {
				return -1;
			}
		} else if ((dellist[i+1] < frame) && (frame < dellist[i+2])) {
			// vaild
			return frame - cutframes;
		}
	}
	// not found
	if (mode < 0) {
		return dellist[dellist.size() - 1];
	}
	return -1;
}


double time2frame(double time)
{
	return time * fps / 1001;	// fps
}

double frame2time(double frame)
{
	return frame * 1001 / fps;	// fps
}


void cut_ass(const char *file_in, const char *file_out)
{
	char fin_path[_MAX_PATH], fout_path[_MAX_PATH];
	strcpy(fin_path, file_in );
	strcat(fin_path, ".ass");
	strcpy(fout_path, file_out );
	strcat(fout_path, ".ass");

	FILE *fpin = fopen(fin_path, "rt");
	if (fpin == NULL) {
		fprintf(stderr, "input ass open error: %s\n", file_in);
		return;
	}
	FILE *fpout = fopen(fout_path, "wt");
	if (fpout == NULL) {
		fprintf(stderr, "output ass open error: %s\n", file_out);
		if (fpin != NULL)	fclose(fpin);
		return;
	}
	char buf[256];
	char *ass_str;
	while (fgets(buf, sizeof(buf), fpin) != NULL) {
		int sh, sm, ss, sms;
		int eh, em, es, ems;
		int Layer;
		if (sscanf(buf, "Dialogue: %d,%01d:%02d:%02d.%02d,%d:%02d:%02d.%02d,",
				&Layer, &sh, &sm, &ss, &sms, &eh, &em, &es, &ems) == 9) {
			ass_str = buf;
			for (int i=0;i<3;i++) {
				ass_str = strchr(ass_str,',');
				ass_str++;
			}
			int start = ((sh*60+sm)*60+ss)*1000+sms*10;		// ass mod
			int end   = ((eh*60+em)*60+es)*1000+ems*10;		// ass mod
			
			start = frame2time(get_cut_frame_number(time2frame(start), +1));
			end = frame2time(get_cut_frame_number(time2frame(end), -1));
			
			if (start >= end) {
				while (fgets(buf, sizeof(buf), fpin) != NULL) {
					if (buf[0] == '\n') {
						break;
					}
				}
				continue;
			}
			
			sms = start % 1000 / 10;
			start /= 1000;
			ss = start % 60;
			start /= 60;
			sm = start % 60;
			start /= 60;
			sh = start % 60;
			
			ems = end % 1000 / 10;
			end /= 1000;
			es = end % 60;
			end /= 60;
			em = end % 60;
			end /= 60;
			eh = end % 60;
			
			fprintf(fpout, "Dialogue: %d,%01d:%02d:%02d.%02d,%d:%02d:%02d.%02d,%s",
					Layer,
					sh, sm, ss, sms, eh, em, es, ems, ass_str);
		} else {
			fputs(buf, fpout);
		}
	}
	fclose(fpout);
	fclose(fpin);
}

void cut_srt(const char *file_in, const char *file_out)
{
	char fin_path[_MAX_PATH], fout_path[_MAX_PATH];
	strcpy(fin_path, file_in );
	strcat(fin_path, ".srt");
	strcpy(fout_path, file_out );
	strcat(fout_path, ".srt");

	int index = 1;
	FILE *fpin = fopen(fin_path, "rt");
	if (fpin == NULL) {
		fprintf(stderr, "input srt open error: %s\n", file_in);
		return;
	}
	FILE *fpout = fopen(fout_path, "wt");
	if (fpout == NULL) {
		fprintf(stderr, "output srt open error: %s\n", file_out);
		if (fpin != NULL)	fclose(fpin);
		return;
	}
//	fprintf(fpout, "\xEF\xBB\xBF");		// write UTF-8 BOM		// mod
	char buf[256];
	while (fgets(buf, sizeof(buf), fpin) != NULL) {
		char *p = buf;
		if (strncmp(buf, "\xEF\xBB\xBF", 3) == 0) {		// skip UTF-8 BOM
			fprintf(fpout, "\xEF\xBB\xBF");		// write UTF-8 BOM		// mod
			p += 3;
		}
		int n;
		if (sscanf(p, "%d\n", &n) != 1) {
			continue;
		}
		if (fgets(buf, sizeof(buf), fpin) == NULL) {
			break;
		}
		
		int sh, sm, ss, sms;
		int eh, em, es, ems;
		if (sscanf(buf, "%d:%02d:%02d,%03d --> %d:%02d:%02d,%03d\n",
				&sh, &sm, &ss, &sms, &eh, &em, &es, &ems) == 8) {
			int start = ((sh*60+sm)*60+ss)*1000+sms;
			int end   = ((eh*60+em)*60+es)*1000+ems;
			
			start = frame2time(get_cut_frame_number(time2frame(start), +1));
			end = frame2time(get_cut_frame_number(time2frame(end), -1));
			
			if (start >= end) {
				while (fgets(buf, sizeof(buf), fpin) != NULL) {
					if (buf[0] == '\n') {
						break;
					}
				}
				continue;
			}
			
			sms = start % 1000;
			start /= 1000;
			ss = start % 60;
			start /= 60;
			sm = start % 60;
			start /= 60;
			sh = start % 60;
			
			ems = end % 1000;
			end /= 1000;
			es = end % 60;
			end /= 60;
			em = end % 60;
			end /= 60;
			eh = end % 60;
			
			fprintf(fpout, "%d\n%02d:%02d:%02d,%03d --> %02d:%02d:%02d,%03d\n",
					index++,
					sh, sm, ss, sms, eh, em, es, ems);
			while (fgets(buf, sizeof(buf), fpin) != NULL) {
				fputs(buf, fpout);
				if (buf[0] == '\n') {
					break;
				}
			}
		} else {
			fputs(buf, fpout);
		}
	}
	fclose(fpout);
	fclose(fpin);
}
